@@ -3,6 +3,7 @@ import asyncio
 import copy
 import logging
 import traceback
+from contextlib import suppress
 from pathlib import Path
 from textwrap import dedent
 from datetime import datetime
@@ -19,6 +20,8 @@ from codearkt.tools import fetch_tools
 from codearkt.event_bus import AgentEventBus, EventType
 from codearkt.llm import LLM, ChatMessages, ChatMessage
 from codearkt.util import get_unique_id, truncate_content
+from codearkt.metrics import TokenUsageStore
+
 
 DEFAULT_END_CODE_SEQUENCE = "<end_code>"
 DEFAULT_END_PLAN_SEQUENCE = "<end_plan>"
@@ -97,6 +100,7 @@ class CodeActAgent:
         self.verbosity_level = verbosity_level
         self.planning_interval = planning_interval
         self.managed_agents: Optional[List[Self]] = managed_agents
+        self.token_usage_store: TokenUsageStore = TokenUsageStore()
 
         if self.managed_agents:
             for agent in self.managed_agents:
@@ -121,8 +125,9 @@ class CodeActAgent:
         messages: ChatMessages,
         session_id: str,
         event_bus: AgentEventBus | None = None,
-        server_host: Optional[str] = None,
-        server_port: Optional[int] = None,
+        token_usage_store: TokenUsageStore | None = None,
+        server_host: str | None = None,
+        server_port: int | None = None,
     ) -> str:
         messages = copy.deepcopy(messages)
 
@@ -174,6 +179,7 @@ class CodeActAgent:
                         run_id=run_id,
                         session_id=session_id,
                         event_bus=event_bus,
+                        token_usage_store=token_usage_store,
                     )
                     messages.extend(new_messages)
                     self._log(
@@ -194,6 +200,7 @@ class CodeActAgent:
                     session_id=session_id,
                     run_id=run_id,
                     event_bus=event_bus,
+                    token_usage_store=token_usage_store,
                     step_number=step_number,
                 )
                 messages.extend(new_messages)
@@ -207,7 +214,11 @@ class CodeActAgent:
             else:
                 # Final step
                 new_messages = await self._handle_final_message(
-                    messages, session_id=session_id, run_id=run_id, event_bus=event_bus
+                    messages,
+                    session_id=session_id,
+                    run_id=run_id,
+                    event_bus=event_bus,
+                    token_usage_store=token_usage_store,
                 )
                 messages.extend(new_messages)
                 self._log(
@@ -256,6 +267,42 @@ class CodeActAgent:
             ), f"Tool {tool_name} not found in {fetched_tool_names}"
         return tools
 
+    async def _run_llm(
+        self,
+        messages: ChatMessages,
+        session_id: str,
+        stop: List[str],
+        event_bus: AgentEventBus | None = None,
+        token_usage_store: TokenUsageStore | None = None,
+        event_type: EventType = EventType.OUTPUT,
+    ) -> str:
+        output_stream = self.llm.astream(messages=messages, stop=stop)
+
+        output_text = ""
+        last_usage = None
+        try:
+            async for event in output_stream:
+                if event.usage:
+                    last_usage = event.usage
+                delta = event.choices[0].delta
+                if isinstance(delta.content, str):
+                    chunk = delta.content
+                elif isinstance(delta.content, list):
+                    chunk = "\n".join([str(item) for item in delta.content])
+                output_text += chunk
+                await self._publish_event(event_bus, session_id, event_type, chunk)
+                if any(ss in output_text for ss in stop):
+                    break
+        finally:
+            with suppress(Exception):
+                await output_stream.aclose()
+        await self._publish_event(event_bus, session_id, event_type, "\n")
+        if last_usage and token_usage_store:
+            await token_usage_store.add(
+                session_id, last_usage.prompt_tokens, last_usage.completion_tokens
+            )
+        return output_text
+
     async def _step(
         self,
         messages: ChatMessages,
@@ -263,6 +310,7 @@ class CodeActAgent:
         session_id: str,
         run_id: str,
         event_bus: AgentEventBus | None = None,
+        token_usage_store: TokenUsageStore | None = None,
         step_number: int | None = None,
     ) -> ChatMessages:
         _ = step_number  # for logs only
@@ -272,21 +320,13 @@ class CodeActAgent:
         self._log("LLM generates outputs...", run_id=run_id, session_id=session_id)
         output_text = ""
         try:
-            output_stream = self.llm.astream(messages, stop=self.prompts.stop_sequences)
-            async for event in output_stream:
-                delta = event.choices[0].delta
-                if isinstance(delta.content, str):
-                    chunk = delta.content
-                elif isinstance(delta.content, list):
-                    chunk = "\n".join([str(item) for item in delta.content])
-                output_text += chunk
-                await self._publish_event(event_bus, session_id, EventType.OUTPUT, chunk)
-                if any(
-                    stop_sequence in output_text for stop_sequence in self.prompts.stop_sequences
-                ):
-                    break
-            await self._publish_event(event_bus, session_id, EventType.OUTPUT, "\n")
-
+            output_text = await self._run_llm(
+                messages=messages,
+                session_id=session_id,
+                stop=self.prompts.stop_sequences,
+                event_bus=event_bus,
+                token_usage_store=token_usage_store,
+            )
         except asyncio.CancelledError:
             raise
         except Exception:
@@ -390,6 +430,7 @@ class CodeActAgent:
         session_id: str,
         run_id: str,
         event_bus: AgentEventBus | None = None,
+        token_usage_store: TokenUsageStore | None = None,
     ) -> ChatMessages:
         prompt: str = self.prompts.final.render()
         final_message = ChatMessage(role="user", content=prompt)
@@ -401,17 +442,13 @@ class CodeActAgent:
             level=logging.DEBUG,
         )
 
-        output_stream = self.llm.astream(input_messages, stop=self.prompts.stop_sequences)
-        output_text = ""
-        async for event in output_stream:
-            delta = event.choices[0].delta
-            if isinstance(delta.content, str):
-                chunk = delta.content
-            elif isinstance(delta.content, list):
-                chunk = "\n".join([str(item) for item in delta.content])
-            output_text += chunk
-            await self._publish_event(event_bus, session_id, EventType.OUTPUT, chunk)
-
+        output_text = await self._run_llm(
+            messages=input_messages,
+            session_id=session_id,
+            stop=self.prompts.stop_sequences,
+            event_bus=event_bus,
+            token_usage_store=token_usage_store,
+        )
         self._log(
             f"Final message: {final_message}",
             run_id=run_id,
@@ -463,6 +500,7 @@ class CodeActAgent:
         run_id: str,
         session_id: str,
         event_bus: AgentEventBus | None = None,
+        token_usage_store: TokenUsageStore | None = None,
     ) -> ChatMessages:
         messages = copy.deepcopy(messages)
         assert self.prompts.plan is not None, "Planning prompt is not set, but planning is enabled"
@@ -481,22 +519,18 @@ class CodeActAgent:
         input_messages = [ChatMessage(role="user", content=planning_prompt)]
 
         try:
-            output_stream = self.llm.astream(input_messages, stop=[self.prompts.end_plan_sequence])
-
             plan_prefix = self.prompts.plan_prefix.render().strip() + "\n\n"
             await self._publish_event(event_bus, session_id, EventType.PLANNING_OUTPUT, plan_prefix)
             output_text = plan_prefix
 
-            async for event in output_stream:
-                delta = event.choices[0].delta
-                chunk = delta.content
-                if chunk is None:
-                    continue
-                assert isinstance(chunk, str), event
-                output_text += chunk
-                await self._publish_event(event_bus, session_id, EventType.PLANNING_OUTPUT, chunk)
-                if self.prompts.end_plan_sequence in output_text:
-                    break
+            output_text += await self._run_llm(
+                messages=input_messages,
+                session_id=session_id,
+                stop=[self.prompts.end_plan_sequence],
+                event_bus=event_bus,
+                token_usage_store=token_usage_store,
+                event_type=EventType.PLANNING_OUTPUT,
+            )
 
             if self.prompts.end_plan_sequence in output_text:
                 output_text = output_text.split(self.prompts.end_plan_sequence)[0].strip()
