@@ -1,6 +1,7 @@
 import os
 import copy
 from typing import Dict, Any, List, cast, AsyncGenerator, Optional
+from contextlib import suppress
 
 from tiktoken import encoding_for_model
 from dotenv import load_dotenv
@@ -45,6 +46,7 @@ class ChatMessage(BaseModel):  # type: ignore
 
 
 ChatMessages = List[ChatMessage]
+ChatStreamGenerator = AsyncGenerator[ChatCompletionChunk, None]
 
 
 def count_openai_tokens(messages: ChatMessages) -> int:
@@ -90,9 +92,7 @@ class LLM:
             messages = first_messages + messages
         return messages
 
-    async def astream(
-        self, messages: ChatMessages, **kwargs: Any
-    ) -> AsyncGenerator[ChatCompletionChunk, None]:
+    async def astream(self, messages: ChatMessages, **kwargs: Any) -> ChatStreamGenerator:
         messages = copy.deepcopy(messages)
         api_params = {**self._params, **kwargs}
 
@@ -124,6 +124,10 @@ class LLM:
                 extra_headers=HEADERS,
                 **api_params,
             )
-            async for event in stream:
-                event_typed: ChatCompletionChunk = event
-                yield event_typed
+            try:
+                async for event in stream:
+                    event_typed: ChatCompletionChunk = event
+                    yield event_typed
+            finally:
+                with suppress(Exception):
+                    await stream.close()
